import os
import shutil
import math
import requests
import json
import datetime
import hashlib
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
import textract
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.specifications.root_document_specification import RootDocumentSpecification
from pygdpr.policies.gdpr_policy import GDPRPolicy
import re
import sys

class Austria(DPA):
    def __init__(self, path=os.curdir):
        country_code='AT'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        from_date = GDPRPolicy().implementation_date().strftime('%d.%m.%Y')
        to_date = datetime.datetime.now().strftime('%d.%m.%Y')
        source = {
            "host": "https://www.ris.bka.gv.at",
            "start_path": f"/Ergebnis.wxe?Abfrage=Dsk&Entscheidungsart=Undefined&Organ=Undefined&SucheNachRechtssatz=True&SucheNachText=True&GZ=&VonDatum=25.05.2018&BisDatum=25.05.2021&Norm=&ImRisSeitVonDatum=&ImRisSeitBisDatum=&ImRisSeit=Undefined&ResultPageSize=100&Suchworte=&Position=1",
        }
        host = source['host']
        start_path = source['start_path']
        if pagination is None or page_soup is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
            return pagination
        pages = page_soup.find('ul', class_='Pages')

        # Removed loop -> place li's into a list and abstract the middle li to get the href
        href_index = 1
        if pages is not None:
            li_list = pages.find_all('li')
            assert li_list
            page_link = li_list[href_index].find('a')
            assert page_link
            page_href = page_link.get('href')
            assert page_href
            pagination.add_item(host + page_href)

        return pagination

    def get_source(self, page_url=None, driver=None, to_print=True):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        pagination = self.update_pagination()
        iteration = 1
        print("\n========================= Austria Decisions Documents ===========================")
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url)
            if page_source is None:
                continue
            page_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert page_soup
            table = page_soup.find('table', class_='bocListTable')
            assert table
            tbody = table.find('tbody', class_='bocListTableBody')
            assert tbody
            # s1. Results
            for tr in tbody.find_all('tr', class_='bocListDataRow'):
                result_index, date_index, document_links_index = 2, 4, 8
                td_list = tr.find_all('td', class_='bocListDataCell')
                assert len(td_list) >= document_links_index + 1
                date_str = td_list[date_index].get_text()
                tmp = datetime.datetime.strptime(date_str, '%d.%m.%Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                result_link = td_list[result_index].find('a')
                assert result_link
                # s2. Documents
                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1
                document_title = result_link.get('title')
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_links = td_list[document_links_index]
                document_href = None
                for document_link in document_links.find_all('a'):
                    cand_href = document_link.get('href')
                    if cand_href.endswith('.pdf'):
                        document_href = cand_href
                        break
                assert document_href
                host = "https://www.ris.bka.gv.at"
                document_url = host + document_href
                if to_print:
                    print('\tdocument_title:\t', document_title)
                    print("\tDocument:\t", document_hash)
                    print('\tdate:\t', str(date))
                    print('\tdocument_url:\t', document_url)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_content = document_response.content
                dpa_folder = self.path

                document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print('Directory path already exists, continue.')

            # s0. Pagination
            pagination = self.update_pagination(pagination=pagination, page_soup=page_soup)
        return added_docs

    # TODO: Get title and date for each document (which will let us create the document_hash as well and filter for date)
    # Using an iterator counter to keep track of documents for now
    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print('------------ GETTING ANNUAL REPORTS ------------')
        added_docs = []

        page_url = "https://www.dsb.gv.at/download-links/dokumente.html"
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url)
        if page_source is None:
            sys.exit("Couldn't obtain page_source from page_url")
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert page_soup

        body = page_soup.find('body', class_='template-article')
        assert body
        main = body.find('main', class_='col-12 content col-xl-9')
        assert main
        span_soup = main.find('span', class_='richtext_output')
        assert span_soup
        ul_list = span_soup.find_all('ul')
        assert ul_list

        iteration = 1
        for ul in ul_list[5:8]:
            for li in ul.find_all('li'):
                assert li

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1;

                # date_str = td_list[date_index].get_text()
                # tmp = datetime.datetime.strptime(date_str, '%d.%m.%Y')
                # date = datetime.date(tmp.year, tmp.month, tmp.day)
                # if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                # continue

                result_link = li.find('a')
                assert result_link

                document_href = result_link.get('href')
                assert document_href

                if ".pdf" in document_href:
                    print("PDF DOC:")
                    document_url = "https://www.dsb.gv.at" + document_href
                    # Check date of pdf link by getting the pdf title
                    pdf_title = result_link.text
                    assert pdf_title
                    document_title = pdf_title

                    # Use re library to look for pdf date in the title string
                    pdf_date_search = re.search(' (.+?) ', pdf_title)
                    if pdf_date_search:
                        found_date = pdf_date_search.group(1)

                    print("PDF year: " + found_date)

                    pdf_end_date = found_date[-4:]
                    print("PDF end date: " + pdf_end_date)

                    if len(pdf_end_date) == 4:
                        if int(pdf_end_date) < 2018:
                            print("\nPDF IS OUTDATED -> SKIPPING\n")
                            continue

                else:
                    print("TEXT DOC:")
                    document_url = document_href
                    document_title = str(iteration)
                print("      Document url:" + document_url)

                # If doc is a pdf, title and date should be correct, if then title is just the iteration number for now
                assert document_title
                print("DOCUMENT TITLE: " + document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                if to_print:
                    print("Document:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                dpa_folder = self.path
                document_folder = dpa_folder + '/' 'AnnualReports' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    if '.pdf' in document_url:
                        document_content = document_response.content

                        with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                            f.write(document_content)
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)

                        added_docs.append(document_hash)
                        iteration += 1
                        continue

                    document_soup = BeautifulSoup(document_response.text, 'html.parser')
                    assert document_soup

                    # Get document_text according to the link scraper is visiting
                    if '.html' in document_url:
                        document_paper = document_soup.find('div', class_='paperw')
                        assert document_paper
                        document_text = document_paper.get_text()
                    elif 'eur-lex.europa.eu' in document_url:
                        document_clear = document_soup.find('div', class_='Wrapper clearfix')
                        assert document_clear
                        document_area = document_clear.find('div', class_='col-md-9')
                        assert document_area
                        document_page = document_area.find('div', class_='tabContent')
                        assert document_page
                        document_text = document_page.get_text()
                    else:
                        document_div = document_soup.find('div', class_='aspNetHidden')
                        assert document_div
                        field_name_body = document_soup.find('div', class_='document', recursive=True)
                        assert field_name_body
                        document_text = field_name_body.get_text()

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title,
                            },
                            'md5': document_hash,
                            'releaseDate': "(Put date here)",  # date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True)
                    added_docs.append(document_hash)  # put doc hash in here

                except FileExistsError:
                    print("Directory path already exists, continue.")
        return added_docs




import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import sys
from datetime import datetime

class EDPB(DPA):
    def __init__(self, path=os.curdir):
        country_code = 'edpb'
        super().__init__(country_code, path)

    # Must include start_path argument on initial call, and new_page_path an subsequent calls
    # start_path should be in the same format as usual
    # new_page_path is 'https://edpb.europa.eu' + start_path + 'unique new page sub-url' + 'new page href'
    # new_page_path is needed because in cases, the 'host' + 'start_path' + 'new page href' doesn't contain
    # the unique new page sub-url
    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path=None, new_page_path=None):
        host = "https://edpb.europa.eu"
        start_path = start_path

        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)

        elif pagination is not None:
            pagination_heading = page_soup.find('ul', class_='pagination')
            assert pagination_heading

            next_page_tag = pagination_heading.find('li', class_='pager__item--next')
            if next_page_tag is None:
                print("\nupdate_pagination: Couldn't find next page button")
            else:
                assert next_page_tag
                page_a = next_page_tag.find('a')
                assert page_a

                page_href = page_a.get('href')
                assert page_href

                page_link = new_page_path + page_href

                # If existing_pages contains the link already, don't add it again
                if pagination.has_link(page_link):
                    pass
                else:
                    print('\nupdate_pagination: adding link: ' + page_link)
                    pagination.add_item(page_link)
        return pagination

    def get_source(self, page_url=None, driver=None, to_print=True):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Recommendations(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guidelines(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Opinions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Letters(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)

        return added_docs



    def get_docs_Recommendations(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= EDPB Recommendations ===========================")
        added_docs = []
        dict_hashcode = {}
        iteration = 1

        page_url = 'https://edpb.europa.eu/our-work-tools/general-guidance/guidelines-recommendations-best-practices_en?f%5B0%5D=opinions_publication_type%3A98'
        if to_print:
            print('\nPage:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("page_source is None")

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        view_content = results_soup.find('div', class_='view-content')
        assert view_content

        view_row_content = view_content.find('div', class_='view-row-content')
        assert view_row_content

        for views_row in view_row_content.find_all('div', class_='views-row'):
            assert views_row

            print('\n------------ Document: ' + str(iteration) + '-------------')
            iteration += 1

            # Get document date
            span = views_row.find('span', class_='news-date')
            assert span

            document_date = span.get_text()
            date_str = datetime.strptime(document_date, '%d %B %Y')
            date_str = date_str.strftime('%d/%m/%Y')
            print('\tDocument Date: ' + date_str)

            # If document year is less than 2018, skip it
            if int(document_date[-4:]) < 2018:
                print("\tSkipping outdated document: " + document_date)
                continue

            # Get document page link
            node_title = views_row.find('h4', class_='node__title')
            assert node_title

            a_tag = node_title.find('a')
            assert a_tag

            document_href = a_tag.get('href')
            assert document_href

            if document_href.startswith('http'):
                document_url = document_href
            else:
                document_url = 'https://edpb.europa.eu' + document_href

            print("\tDocument Page URL: " + document_url)

            # Get document title
            node_span = node_title.find('span')
            assert node_span

            document_title = node_span.get_text()
            print('\tDocument Title: ' + document_title)

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            # documents have the same hashcode, but different dates
            date_part = date_str.replace('/', '_')
            if document_hash in dict_hashcode:
                document_hash = document_hash + '-' + date_part

            if to_print:
                print("\tDocument:\t", document_hash)

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            article = document_soup.find('article', role='article')
            assert article

            # Check if there is 'final document version' notice that has a link -> if so, use this link
            alert_document = article.find('div', class_='alert')
            if alert_document is not None:
                print('\tGetting final version after public consultation')
                alert_a = alert_document.find('a')
                assert alert_a

                alert_href = alert_a.get('href')
                assert alert_href

                if alert_href.startswith('http'):
                    alert_url = alert_href
                else:
                    alert_url = 'https://edpb.europa.eu' + alert_href

                print('alert_url:' + alert_url)
                # Visit the page that contains the download link for the pdf
                alert_response = None
                try:
                    alert_response = requests.request('GET', alert_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if alert_response is None:
                    continue

                # Now get the pdf download link
                alert_soup = BeautifulSoup(alert_response.text, 'html.parser')
                assert alert_soup

                alert_article = alert_soup.find('article', role='article')
                assert alert_article

                col_sm_2 = alert_article.find('div', class_='col-sm-2')
                assert col_sm_2

                document_a = col_sm_2.find('a')
                assert document_a

                pdf_href = document_a.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://edpb.europa.eu' + pdf_href

                print('\tPDF URL: ' + pdf_url)

                # Download the pdf
                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

            else:
                col_sm_2 = article.find('div', class_='col-sm-2')
                assert col_sm_2

                document_a = col_sm_2.find('a')
                assert document_a

                pdf_href = document_a.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://edpb.europa.eu' + pdf_href

                print('\tPDF URL: ' + pdf_url)

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

            dpa_folder = self.path

            dpa_folder = self.path
            # document_folder = dpa_folder + '/edpb' + '/' + 'Recommendations' + '/' + document_hash
            document_folder = dpa_folder + '/' + 'Recommendations' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(pdf_response.content)
                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    try:
                        pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(pdf_text)
                    except:
                        pass
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': date_str,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True)
                added_docs.append(document_hash)
                dict_hashcode[document_hash] = date_part
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs

    def get_docs_Guidelines(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= EDPB Guidelines ===========================")
        added_docs = []
        dict_hashcode = {}
        # Starting url is set to page 1, instead of the generic url (which lacks page specifier but starts at page 1)
        pagination = self.update_pagination(start_path='/our-work-tools/general-guidance/guidelines-recommendations-best-practices_en?f%5B0%5D=opinions_publication_type%3A64')

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\nPage:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                print("page_source is None")
                continue

            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            view_content = results_soup.find('div', class_='view-content')
            assert view_content

            view_row_content = view_content.find('div', class_='view-row-content')
            assert view_row_content

            for views_row in view_row_content.find_all('div', class_='views-row'):
                assert views_row

                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                # Get document date
                span = views_row.find('span', class_='news-date')
                assert span

                document_date = span.get_text()
                date_str = datetime.strptime(document_date, '%d %B %Y')
                date_str = date_str.strftime('%d/%m/%Y')
                print('\tDocument date: ' + date_str)

                # If document year is less than 2018, skip it
                if int(document_date[-4:]) < 2018:
                    print("\tSkipping outdated document: " + document_date)
                    continue

                # Get document page link
                node_title = views_row.find('h4', class_='node__title')
                assert node_title

                a_tag = node_title.find('a')
                assert a_tag

                document_href = a_tag.get('href')
                assert document_href

                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://edpb.europa.eu' + document_href

                print("\tDocument Page URL: " + document_url)

                # Get document title
                node_span = node_title.find('span')
                assert node_span

                document_title = node_span.get_text()
                print('\tDocument Title: ' + document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                # documents have the same hashcode, but different dates
                date_part = date_str.replace('/', '_')
                if document_hash in dict_hashcode:
                    document_hash = document_hash + '-' + date_part
                print('\tDocument hash: ' + document_hash)

                dpa_folder = self.path

                # document_folder = dpa_folder + '/edpb' + '/' + 'Guidelines' + '/' + document_hash
                document_folder = dpa_folder + '/' + 'Guidelines' + '/' + document_hash

                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                article = document_soup.find('article', role='article')
                assert article

                # If the standard download button is missing
                if article.find('div', class_='col-sm-2') is None:
                    print('Standard download button is missing')

                    # avoid broken links
                    broken_links = ['bcce02dab0081e0017b2d4709b0d7df3', '1725b5ab1d3cf73aac4c82cb0246d717']
                    if document_hash in broken_links:
                        print('links is not working, continue')
                        continue

                    field_item = article.find('div', class_='field-item')

                    if field_item is None:
                        field_item = article.find('div', class_='text-formatted')
                        print('doc 5 should be here')

                    field_a = field_item.find('a')
                    assert field_a

                    field_href = field_a.get('href')
                    assert field_href

                    field_response = None
                    try:
                        field_response = requests.request('GET', field_href)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if field_response is None:
                        continue

                    # Now get the pdf download links
                    field_soup = BeautifulSoup(field_response.text, 'html.parser')
                    assert field_soup

                    cnect = field_soup.find('div', class_='cnect-main-body')
                    assert cnect

                    try:
                        os.makedirs(document_folder)

                        pdf_iterator = 1
                        cnect_list = cnect.find_all('div', class_='ecl-u-mt-m')
                        # type 1 document
                        if len(cnect_list) == 0:
                            a = cnect.find('a')
                            href = a.get('href')
                            pdf_response = None
                            try:
                                pdf_response = requests.request('GET', href)
                            except requests.exceptions.HTTPError as error:
                                if to_print:
                                    print(error)
                                pass
                            if pdf_response is None:
                                continue

                            with open(document_folder + '/' + self.language_code + '.pdf',
                                      'wb') as f:
                                f.write(pdf_response.content)
                            with open(document_folder + '/' + self.language_code + '.txt',
                                      'wb') as f:
                                try:
                                    pdf_text = textract.process(
                                        document_folder + '/' + self.language_code + '.pdf')
                                    f.write(pdf_text)
                                except:
                                    pass
                        # type 2 document
                        else:
                            for div in cnect_list:

                                assert div
                                a = div.find('a')
                                assert a

                                href = a.get('href')
                                assert href

                                pdf_response = None
                                try:
                                    pdf_response = requests.request('GET', href)
                                except requests.exceptions.HTTPError as error:
                                    if to_print:
                                        print(error)
                                    pass
                                if pdf_response is None:
                                    continue

                                with open(document_folder + '/' + self.language_code + str(pdf_iterator) + '.pdf', 'wb') as f:
                                    f.write(pdf_response.content)
                                with open(document_folder + '/' + self.language_code + str(pdf_iterator) + '.txt', 'wb') as f:
                                    try:
                                        pdf_text = textract.process(document_folder + '/' + self.language_code + str(pdf_iterator) + '.pdf')
                                        f.write(pdf_text)
                                    except:
                                        pass
                                pdf_iterator += 1
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date_str,
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True)
                        added_docs.append(document_hash)
                        dict_hashcode[document_hash] = date_part

                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")

                    # Move to next outer for loop iteration (move to next overall document page on e)
                    continue

                # Check if there is 'final document version' notice that has a link -> if so, use this link
                alert_document = article.find('div', class_='alert')
                if alert_document is not None:
                    print('\tGetting final version after public consultation')
                    alert_a = alert_document.find('a')
                    assert alert_a

                    alert_href = alert_a.get('href')
                    assert alert_href

                    if alert_href.startswith('http'):
                        alert_url = alert_href
                    else:
                        alert_url = 'https://edpb.europa.eu' + alert_href

                    # Visit the page that contains the download link for the pdf
                    alert_response = None
                    try:
                        alert_response = requests.request('GET', alert_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if alert_response is None:
                        continue

                    # Now get the pdf download link
                    alert_soup = BeautifulSoup(alert_response.text, 'html.parser')
                    assert alert_soup

                    alert_article = alert_soup.find('article', role='article')
                    assert alert_article

                    col_sm_2 = alert_article.find('div', class_='col-sm-2')
                    assert col_sm_2

                    document_a = col_sm_2.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    # Download the pdf
                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                # There is no 'updated document after public consultation' link
                else:
                    col_sm_2 = article.find('div', class_='col-sm-2')
                    assert col_sm_2

                    document_a = col_sm_2.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        try:
                            pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(pdf_text)
                        except:
                            pass
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True)
                    added_docs.append(document_hash)
                    dict_hashcode[document_hash] = date_part
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup, new_page_path='https://edpb.europa.eu/our-work-tools/general-guidance/guidelines-recommendations-best-practices_en')
        return added_docs

    def get_docs_Opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= EDPB Opinions ===========================")
        added_docs = []
        # Starting url is set to page 1, instead of the generic url (which lacks page specifier but starts at page 1)
        pagination = self.update_pagination(start_path='/our-work-tools/consistency-findings/opinions_en')

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\nPage:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                print("page_source is None")
                continue

            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            view_content = results_soup.find('div', class_='view-content')
            assert view_content

            view_row_content = view_content.find('div', class_='view-row-content')
            assert view_row_content

            for views_row in view_row_content.find_all('div', class_='views-row'):
                assert views_row

                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                # Get document date
                span = views_row.find('span', class_='news-date')
                assert span

                document_date = span.get_text()
                date_str = datetime.strptime(document_date, '%d %B %Y')
                date_str = date_str.strftime('%d/%m/%Y')
                print('\tDocument date: ' + date_str)

                # If document year is less than 2018, skip it
                if int(document_date[-4:]) < 2018:
                    print("\tSkipping outdated document: " + document_date)
                    continue

                # Get document page link
                node_title = views_row.find('h4', class_='node__title')
                assert node_title

                a_tag = node_title.find('a')
                assert a_tag

                document_href = a_tag.get('href')
                assert document_href

                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://edpb.europa.eu' + document_href

                print("\tDocument Page URL: " + document_url)

                # Get document title
                node_span = node_title.find('span')
                assert node_span

                document_title = node_span.get_text()
                print('\tDocument Title: ' + document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                if to_print:
                    print("\tDocument:\t", document_hash)

                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                article = document_soup.find('article', role='article')
                assert article

                # Check if there is 'final document version' notice that has a link -> if so, use this link
                alert_document = article.find('div', class_='alert')
                if alert_document is not None:
                    print('\tGetting final version after public consultation')
                    alert_a = alert_document.find('a')
                    assert alert_a

                    alert_href = alert_a.get('href')
                    assert alert_href

                    if alert_href.startswith('http'):
                        alert_url = alert_href
                    else:
                        alert_url = 'https://edpb.europa.eu' + alert_href

                    # Visit the page that contains the download link for the pdf
                    alert_response = None
                    try:
                        alert_response = requests.request('GET', alert_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if alert_response is None:
                        continue

                    # Now get the pdf download link
                    alert_soup = BeautifulSoup(alert_response.text, 'html.parser')
                    assert alert_soup

                    alert_article = alert_soup.find('article', role='article')
                    assert alert_article

                    col_sm_2 = alert_article.find('div', class_='col-sm-2')
                    assert col_sm_2

                    document_a = col_sm_2.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    # Download the pdf
                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                # There is no 'updated document after public consultation' link
                else:
                    col_sm_2 = article.find('div', class_='col-sm-2')
                    assert col_sm_2

                    document_a = col_sm_2.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    except requests.exceptions.ConnectionError as error:
                        if to_print:
                            print(error)
                    if pdf_response is None:
                        continue

                dpa_folder = self.path
                # document_folder = dpa_folder + '/edpb' + '/' + 'Opinions' + '/' + document_hash
                document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        try:
                            pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(pdf_text)
                        except:
                            pass
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup, new_page_path='https://edpb.europa.eu/our-work-tools/consistency-findings/opinions_en')
        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= EDPB Decision Documents ===========================")
        added_docs = []

        iteration = 1

        page_url = 'https://edpb.europa.eu/our-work-tools/consistency-findings/binding-decisions_en'
        if to_print:
            print('\nPage:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("page_source is None")

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        view_content = results_soup.find('div', class_='view-content')
        assert view_content

        view_row_content = view_content.find('div', class_='view-row-content')
        assert view_row_content

        for views_row in view_row_content.find_all('div', class_='views-row'):
            assert views_row

            print('\n------------ Document: ' + str(iteration) + '-------------')
            iteration += 1

            # Get document date
            span = views_row.find('span', class_='news-date')
            assert span

            document_date = span.get_text()
            date_str = datetime.strptime(document_date, '%d %B %Y')
            date_str = date_str.strftime('%d/%m/%Y')
            print('\tDocument Date: ' + date_str)

            # If document year is less than 2018, skip it
            if int(document_date[-4:]) < 2018:
                print("\tSkipping outdated document: " + document_date)
                continue

            # Get document page link
            node_title = views_row.find('h4', class_='node__title')
            assert node_title

            a_tag = node_title.find('a')
            assert a_tag

            document_href = a_tag.get('href')
            assert document_href

            if document_href.startswith('http'):
                document_url = document_href
            else:
                document_url = 'https://edpb.europa.eu' + document_href

            print("\tDocument Page URL: " + document_url)

            # Get document title
            node_span = node_title.find('span')
            assert node_span

            document_title = node_span.get_text()
            print('\tDocument Title: ' + document_title)

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            if to_print:
                print("\tDocument:\t", document_hash)

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            article = document_soup.find('article', role='article')
            assert article

            # Check if there is 'final document version' notice that has a link -> if so, use this link
            alert_document = article.find('div', class_='alert')
            if alert_document is not None:
                print('\tGetting final version after public consultation')
                alert_a = alert_document.find('a')
                assert alert_a

                alert_href = alert_a.get('href')
                assert alert_href

                if alert_href.startswith('http'):
                    alert_url = alert_href
                else:
                    alert_url = 'https://edpb.europa.eu' + alert_href

                print('alert_url:' + alert_url)
                # Visit the page that contains the download link for the pdf
                alert_response = None
                try:
                    alert_response = requests.request('GET', alert_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if alert_response is None:
                    continue

                # Now get the pdf download link
                alert_soup = BeautifulSoup(alert_response.text, 'html.parser')
                assert alert_soup

                alert_article = alert_soup.find('article', role='article')
                assert alert_article

                col_sm_2 = alert_article.find('div', class_='col-sm-2')
                assert col_sm_2

                document_a = col_sm_2.find('a')
                assert document_a

                pdf_href = document_a.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://edpb.europa.eu' + pdf_href

                print('\tPDF URL: ' + pdf_url)

                # Download the pdf
                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

            else:
                col_sm_2 = article.find('div', class_='col-sm-2')
                assert col_sm_2

                document_a = col_sm_2.find('a')
                assert document_a

                pdf_href = document_a.get('href')
                assert pdf_href

                if pdf_href.startswith('http'):
                    pdf_url = pdf_href
                else:
                    pdf_url = 'https://edpb.europa.eu' + pdf_href

                print('\tPDF URL: ' + pdf_url)

                pdf_response = None
                try:
                    pdf_response = requests.request('GET', pdf_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

            dpa_folder = self.path
            # document_folder = dpa_folder + '/edpb' + '/' + 'Decisions' + '/' + document_hash
            document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(pdf_response.content)
                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    try:
                        pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(pdf_text)
                    except:
                        pass
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': date_str,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs

    # Some old document pages are blank -> skip them
    def get_docs_Letters(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= EDPB Letters ===========================")
        added_docs = []

        pagination = self.update_pagination(start_path='/our-work-tools/documents/letters_en')

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\nPage:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                print("page_source is None")
                continue

            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            view_content = results_soup.find('div', class_='view-content')
            assert view_content

            view_row_content = view_content.find('div', class_='view-row-content')
            assert view_row_content

            for views_row in view_row_content.find_all('div', class_='views-row'):
                assert views_row

                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                # Get document date
                span = views_row.find('span', class_='news-date')
                assert span

                document_date = span.get_text()
                date_str = datetime.strptime(document_date, '%d %B %Y')
                date_str = date_str.strftime('%d/%m/%Y')
                print('\tDocument Date: ' + date_str)

                # If document year is less than 2018, skip it
                if int(document_date[-4:]) < 2018:
                    print("\tSkipping outdated document: " + document_date)
                    continue

                # Get document page link
                node_title = views_row.find('h4', class_='node__title')
                assert node_title

                a_tag = node_title.find('a')
                assert a_tag

                document_href = a_tag.get('href')
                assert document_href

                if document_href.startswith('http'):
                    document_url = document_href
                else:
                    document_url = 'https://edpb.europa.eu' + document_href

                print("\tDocument Page URL: " + document_url)

                # Get document title
                node_span = node_title.find('span')
                assert node_span

                document_title = node_span.get_text()
                print('\tDocument Title: ' + document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                if to_print:
                    print("\tDocument:\t", document_hash)

                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                article = document_soup.find('article', role='article')
                assert article

                # Check if there is 'final document version' notice that has a link -> if so, use this link
                alert_document = article.find('div', class_='alert')
                if alert_document is not None:
                    print('\tGetting final version after public consultation')
                    alert_a = alert_document.find('a')
                    assert alert_a

                    alert_href = alert_a.get('href')
                    assert alert_href

                    if alert_href.startswith('http'):
                        alert_url = alert_href
                    else:
                        alert_url = 'https://edpb.europa.eu' + alert_href

                    # Visit the page that contains the download link for the pdf
                    alert_response = None
                    try:
                        alert_response = requests.request('GET', alert_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if alert_response is None:
                        continue

                    # Now get the pdf download link
                    alert_soup = BeautifulSoup(alert_response.text, 'html.parser')
                    assert alert_soup

                    alert_article = alert_soup.find('article', role='article')
                    assert alert_article

                    col_sm_2 = alert_article.find('div', class_='col-sm-2')
                    assert col_sm_2

                    document_a = col_sm_2.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    # Download the pdf
                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                # There is no 'updated document after public consultation' link
                else:
                    col_sm_2 = article.find('div', class_='col-sm-2')

                    # This usually means the document page is blank
                    if col_sm_2 is None:
                        print("Unable to parse data on document page")
                        continue

                    document_a = col_sm_2.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                dpa_folder = self.path
                # document_folder = dpa_folder + '/edpb' + '/' + 'Letters' + '/' + document_hash
                document_folder = dpa_folder + '/' + 'Letters' + '/' + document_hash

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        try:
                            pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(pdf_text)
                        except:
                            pass
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup, new_page_path='https://edpb.europa.eu/our-work-tools/documents/letters_en')
        return added_docs

    # TODO: Using this method to test the document hash checking currently
    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):

        print("\n========================= EDPB Annual Reports ===========================")
        added_docs = []

        iteration = 1

        page_url = 'https://edpb.europa.eu/about-edpb/about-edpb/annual-reports_en'
        if to_print:
            print('\nPage:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("page_source is None")

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        view_content = results_soup.find('div', class_='view-content')
        assert view_content

        view_row_content = view_content.find('div', class_='view-row-content')
        assert view_row_content

        for views_row in view_row_content.find_all('div', class_='views-row'):
            assert views_row

            print('\n------------ Document: ' + str(iteration) + '-------------')
            iteration += 1

            # Get document date
            span = views_row.find('span', class_='news-date')
            assert span

            document_date = span.get_text()
            date_str = datetime.strptime(document_date, '%d %B %Y')
            date_str = date_str.strftime('%d/%m/%Y')
            print('\tDocument Date: ' + date_str)

            # If document year is less than 2018, skip it
            if int(document_date[-4:]) < 2018:
                print("\tSkipping outdated document: " + document_date)
                continue

            # Get document page link
            node_title = views_row.find('h4', class_='node__title')
            assert node_title

            a_tag = node_title.find('a')
            assert a_tag

            document_href = a_tag.get('href')
            assert document_href

            if document_href.startswith('http'):
                document_url = document_href
            else:
                document_url = 'https://edpb.europa.eu' + document_href

            print("\tDocument Page URL: " + document_url)

            # Get document title
            node_span = node_title.find('span')
            assert node_span

            document_title = node_span.get_text()
            print('\tDocument Title: ' + document_title)

            # TODO: Included url in doc hash -> then use global list for doc hashes
            document_hash = hashlib.md5(document_title.encode()).hexdigest()

            if (document_hash in existing_docs) and (overwrite is False):
                if to_print:
                    print('\tSkipping existing document (It exists in visitedDocs.txt):\t', document_hash)
                continue

            if to_print:
                print("\tDocument:\t", document_hash)

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            article = document_soup.find('article', role='article')
            assert article

            col_sm_2 = article.find_all('div', class_='col-sm-2')
            assert col_sm_2

            dpa_folder = self.path
            # Put documents in folder for the dpa automatically
            # document_folder = dpa_folder + '/edpb' + '/' + 'Annual Reports' + '/' + document_hash
            document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                # There are usually two pdf's -> the actual report, and an executive summary
                pdf_iteration = 1
                for col in col_sm_2:
                    document_a = col.find('a')
                    assert document_a

                    pdf_href = document_a.get('href')
                    assert pdf_href

                    if pdf_href.startswith('http'):
                        pdf_url = pdf_href
                    else:
                        pdf_url = 'https://edpb.europa.eu' + pdf_href

                    print('\tPDF URL: ' + pdf_url)

                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url)
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                    with open(document_folder + '/' + self.language_code + str(pdf_iteration) + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)
                    with open(document_folder + '/' + self.language_code + str(pdf_iteration) + '.txt', 'wb') as f:
                        try:
                            pdf_text = textract.process(document_folder + '/' + self.language_code + str(pdf_iteration) + '.pdf')
                            f.write(pdf_text)
                        except:
                            pass
                    pdf_iteration += 1

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': date_str,
                        'releaseDate': document_date,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        # Contains unique hashes not in master file...
        return added_docs



import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import sys
import time

# TODO: Implement Luxembourg guidelines scraper
class Luxembourg(DPA):
    def __init__(self, path=os.curdir):
        country_code='lu'
        super().__init__(country_code, path)

    # Passing this method a page parse object causes it to find ALL links to next page buttons visible at bottom of page
    # The visible page buttons only change every 5 page flips, so need to check that duplicates aren't being added
    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path="Decisions"):
        source = {
            "host": "https://cnpd.public.lu",
            "start_path_opinion": "/fr/decisions-avis.html?b=0",
            "start_path_decision": "/fr/decisions-sanctions.html?b=0"
        }
        if start_path != "Decisions":
            start_path = source['start_path_opinion']
        else:
            start_path = source['start_path_decision']
        host = source['host']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        else:
            ol_pagination = page_soup.find('ol', class_='pagination')
            assert ol_pagination
            for li in ol_pagination.find_all('li', class_='pagination-page'):
                page_link = li.find('a')
                if page_link is None:
                    continue
                page_href = page_link.get('href')
                page_link = host + page_href

                # If existing_pages contains the link already, don't add it again
                if pagination.has_link(page_link):
                    continue
                print("\nLINK ADDED TO PAGINATION OBJECT: " + host + page_href)
                pagination.add_item(page_link)
        return pagination

    def get_source(self, page_url=None, driver=None, to_print=True):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Opinions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Luxembourg Decision Documents =========================")
        added_docs = []
        # Starting url is set to page 1, instead of the generic url (which lacks page specifier but starts at page 1)
        pagination = self.update_pagination()

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\nPage:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            search_results = results_soup.find('ol', class_='search-results')
            assert search_results
            # s1. Results
            for li in search_results.find_all('li', recursive=False):
                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1
                time.sleep(2)

                article_time = li.find('time', class_='article-published')
                assert article_time
                date_str = article_time.get('datetime')
                date_str = date_str.split()[0]

                tmp = datetime.datetime.strptime(date_str, '%Y-%m-%d')
                date = datetime.date(tmp.year, tmp.month, tmp.day)


                # If document year is less than 2018, skip it
                if int(date_str[0:4]) < 2018:
                    print("\tSkipping outdated document: " + date.strftime('%d/%m/%Y'))
                    return added_docs

                print('\tDocument Date: ' + date.strftime('%d/%m/%Y'))

                '''
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    print("\tSkipping outdated document: " + date.strftime('%d/%m/%Y'))
                    continue
                '''

                article_title = li.find('h2', class_='article-title')
                assert article_title
                result_link = article_title.find('a')
                assert result_link

                document_title = result_link.get_text()
                print('\tDocument Title:\t' + document_title)
                print('\tdate:\t', date)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://cnpd.public.lu"
                document_url = host + document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup
                aside = document_soup.find('aside', class_='page-more')
                assert aside
                file_links = aside.find_all('a')
                file_url = None
                for file_link in file_links:
                    file_href = file_link.get('href')
                    if file_href.endswith('.pdf'):
                        if file_href.startswith('http') is False:
                            file_url = host + file_href
                        else:
                            file_url = file_href
                        break
                file_response = None
                try:
                    print('\tfile_url:', file_url)
                    file_response = requests.request('GET', file_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if file_response is None:
                    continue
                file_content = file_response.content
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(file_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        try:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                        except:
                            pass
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': file_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            # s0. Pagination
            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup)

        return added_docs

    def get_docs_Opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Luxembourg Opinions =========================")
        added_docs = []
        # Starting url is set to page 1, instead of the generic url (which lacks page specifier but starts at page 1)
        pagination = self.update_pagination(start_path = "Opinions")

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\nPage:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            search_results = results_soup.find('ol', class_='search-results')
            assert search_results
            # s1. Results
            for li in search_results.find_all('li', recursive=False):
                print('\n------------ Document: ' + str(iteration) + '-------------')
                iteration += 1

                article_time = li.find('time', class_='article-published')
                assert article_time
                date_str = article_time.get('datetime')
                date_str = date_str.split()[0]

                tmp = datetime.datetime.strptime(date_str, '%Y-%m-%d')
                date = datetime.date(tmp.year, tmp.month, tmp.day)

                # If document year is less than 2018, skip it
                if int(date_str[0:4]) < 2018:
                    print("\tSkipping outdated document: " + date.strftime('%d/%m/%Y'))
                    return added_docs

                print('\tDocument Date: ' + date.strftime('%d/%m/%Y'))

                '''
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    print("\tSkipping outdated document: " + date.strftime('%d/%m/%Y'))
                    continue
                '''

                article_title = li.find('h2', class_='article-title')
                assert article_title
                result_link = article_title.find('a')
                assert result_link

                document_title = result_link.get_text()
                print('\tDocument Title:\t' + document_title)
                print('\tdate_str:\t', date_str)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://cnpd.public.lu"
                document_url = host + document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup
                aside = document_soup.find('aside', class_='page-more')
                assert aside
                file_links = aside.find_all('a')
                file_url = None
                for file_link in file_links:
                    file_href = file_link.get('href')
                    if file_href.endswith('.pdf'):
                        if file_href.startswith('http') is False:
                            file_url = host + file_href
                        else:
                            file_url = file_href
                        break
                file_response = None
                try:
                    print('\tfile_url:', file_url)
                    file_response = requests.request('GET', file_url)
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if file_response is None:
                    continue
                file_content = file_response.content
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(file_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        try:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                        except:
                            pass
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': file_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            # s0. Pagination
            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup, start_path="Opinions")

        return added_docs

    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Luxembourg Annual Reports =========================")
        added_docs = []

        iteration = 1

        page_url = 'https://cnpd.public.lu/fr/publications/rapports.html'
        if to_print:
            print('\nPage:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("Couldn't connect to base url")

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        page_text = results_soup.find('div', class_='page-text')
        assert page_text

        ul = page_text.find('ul')
        assert ul

        for li in ul.find_all('li'):
            assert li

            print('\n------------ Document: ' + str(iteration) + '-------------')
            iteration += 1

            result_link = li.find('a')
            assert result_link

            document_title = result_link.get_text()
            print('\tDocument Title:\t' + document_title)

            # Check if document is outdated
            document_date = document_title[-4:]
            print('\tDocument Date:\t' + document_date)
            if int(document_date) < 2018:
                print('\tSkipping outdated document: ' + document_date)
                return added_docs

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = result_link.get('href')
            assert document_href
            host = "https:"
            document_url = host + document_href
            if to_print:
                print("\tDocument:\t", document_hash)

            print("\tDocument URL: " + document_url)
            document_response = None
            try:
                document_response = requests.request('GET', document_url)
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash
            try:
                os.makedirs(document_folder)
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_response.content)
                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    try:
                        document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                    except:
                        print('\tFailed to convert pdf to text document')
                        pass
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_date,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs


import nltk

class PreprocessWordsService():
    def for_words(self, words):
        if len(words) == 0:
            return []
        stop_words = set(nltk.corpus.stopwords.words('english'))
        words = [w for w in words if w.isalpha()]
        words = [w.lower() for w in words]
        words = [w for w in words if w not in stop_words]
        return words



import os
import io
import shutil
#import pytesseract
from pdfminer.converter import TextConverter
from pdfminer.converter import PDFPageAggregator

from pdfminer.pdfinterp import PDFPageInterpreter
from pdfminer.pdfinterp import PDFResourceManager

from pdfminer.pdfpage import PDFPage
from pdfminer.layout import LTImage, LTFigure

from pdf2image import convert_from_path
from PIL import Image # brew install pillow
from pytesseract import image_to_string # brew install tesseract

from ..specifications.pdf_file_extension_specification import PDFFileExtensionSpecification
from .filename_from_path_service import filename_from_path_service

TMP_PATH = '/tmp'

#pytesseract.pytesseract.tesseract_cmd = r'/usr/local/Cellar/tesseract/4.1.1/bin/tesseract'
class PDFToTextService:
    # subroutine
    def text_from_pdf_images_path(self, path):
        if PDFFileExtensionSpecification().is_satisfied_by(path) is False:
            raise ValueError("path is not of type pdf")

        #pages = convert_from_path(path, 500, poppler_path=r'/usr/local/Cellar/poppler/21.09.0/bin')
        pages = convert_from_path(path, 500)

        if len(pages) == 0:
            return None

        filename = filename_from_path_service(path)
        dirpath = TMP_PATH + '/' + filename

        try:
            os.makedirs(dirpath)
            #os.mkdir(dirpath)
        except FileExistsError:
            print("directory already exists.")

        text = ''
        for i in range(len(pages)):
            im_path = '{path}/{no}.jpg'.format(path=dirpath, no=i)

            page = pages[i]
            page.save(im_path, 'JPEG')

            im = Image.open(im_path)
            text += image_to_string(im, lang='eng')

            if i != len(pages) - 1:
                text += '\n'

        shutil.rmtree(dirpath, ignore_errors=True)

        text = text.strip()
        return text

    def text_from_pdf_path(self, path):
        if PDFFileExtensionSpecification().is_satisfied_by(path) is False:
            raise ValueError("path is not of type pdf")

        resource_manager = PDFResourceManager()
        fake_file_handle = io.StringIO()
        converter = TextConverter(resource_manager, fake_file_handle)
        page_interpreter = PDFPageInterpreter(resource_manager, converter)

        with open(path, 'rb') as fh:
            for page in PDFPage.get_pages(fh,
                                          caching=True,
                                          check_extractable=True):
                page_interpreter.process_page(page)
            text = fake_file_handle.getvalue()

        # close open handles
        converter.close()
        fake_file_handle.close()

        text = text.strip()

        if len(text) == 0:
            text = self.text_from_pdf_images_path(path)

        return text



import os
import json

class ListEUCurrenciesService():
    def get(self):
        eu_currencies = {}
        path = os.path.abspath('pygdpr/assets/eu-currencies.json')
        f = open(path, 'r')
        eu_currencies_items = json.load(f)
        f.close()
        for country_code, currency in eu_currencies_items.items():
            currency_code = currency['code'].upper()
            if currency_code not in eu_currencies.keys():
                eu_currencies[currency_code] = currency
        return list(eu_currencies.values())



import os
import sys
import time
import shutil
import math
import requests
import json
import datetime
import hashlib
import textract
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
import dateparser
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy

from zipfile import ZipFile


class Belgium(DPA):
    def __init__(self, path=os.curdir):
        country_code='BE'
        super().__init__(country_code, path)

    # Modified method to take the host and start path as inputs
    def update_pagination(self, pagination=None, page_soup=None, host_link_input=None, start_path_input=None):
        #if pagination is None:
        pagination = Pagination()
        pager = page_soup.find('ul', class_='pagination')
        page_repeat = False
        if pager is not None:
            for li in pager.find_all('li', class_='page-item'):
                page_link = li.find('a')
                if page_link is None:
                    continue
                host = "https://www.autoriteprotectiondonnees.be"
                page_href = page_link.get('href')
                # avoid duplicate page contents
                if page_repeat and '&l=50&p=1' in page_href:
                    return pagination
                if '&l=50&p=1' in page_href:
                    page_repeat = True
                page_url = host + page_href
                # print(page_url)
                pagination.add_item(page_url)
        else:
            page_url = host_link_input + start_path_input
            pagination.add_item(page_url)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url, timeout=1000)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    # Calls all scraper methods at once
    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions_v1(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Decisions_v2(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Opinions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guides(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    # TODO: Review this decision to put both methods for decisions under one single method to be called in CLI
    # Calls both decisions scrapers and returns all document hashes collected
    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        added_docs += self.get_docs_Decisions_v1()
        added_docs += self.get_docs_Decisions_v2()

        return added_docs

    # Gets all documents located at first Decisions link
    # Date checking correct
    def get_docs_Decisions_v1(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []

        host_link_input = "https://www.autoriteprotectiondonnees.be"
        start_path_input = "/citoyen/chercher?q=&search_category%5B%5D=taxonomy%3Apublications&search_type%5B%5D=decision&search_subtype%5B%5D=taxonomy%3Adispute_chamber_substance_decisions&s=recent&l=50"
        iteration_number = 1
        page_url = host_link_input + start_path_input
        page_source = self.get_source(page_url)
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        pagination = self.update_pagination(page_soup=page_soup)
        print("\n========================= Belgium Decisions_v1 Documents ===========================")

        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url)
            page_soup = BeautifulSoup(page_source.text, 'html.parser')
            search_result = page_soup.find('div', id='search-result')
            assert search_result
            # s1. Results
            for media in search_result.find_all('div', class_='media'):
                time.sleep(1)
                media_title = media.find('h3', class_='media-title')
                print("------------ Document " + str(iteration_number) + " ------------")
                iteration_number += 1
                assert media_title
                result_link = media_title.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://www.autoriteprotectiondonnees.be"
                document_url = host + document_href
                print('document_url:', document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=1000)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    if document_url.endswith('.pdf') is False:
                        document_soup = BeautifulSoup(document_response.text, 'html.parser')
                        assert document_soup
                        date_text = document_soup.find('div', class_='date').get_text()
                        date_str = date_text[-4:] # date_text[:-4] + ' ' + date_text[-4:]
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        date = datetime.date(tmp.year, tmp.month, tmp.day)
                        if date.year < 2018:
                            print("Skipping outdated document")
                            continue
                        page_body = document_soup.find('div', class_='page-body')
                        assert page_body
                        document_text = page_body.get_text()
                    else:
                        date_str = document_title.split(' du ')[-1]
                        #print("date_str:", date_str)
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        if tmp is None:
                            media_date = media.find('span', class_="media-date")
                            assert media_date
                            year = int(media_date.get_text())
                            if year < 2018:
                                continue
                        else:
                            # Use a fixed date (GDRP release date) rather than a moving window
                            date = datetime.date(tmp.year, tmp.month, tmp.day)
                            #if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                                #continue
                            if date.year < 2018:
                                print("Skipping outdated document")
                                continue

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_response.content)
                    if document_url.endswith('.pdf'):
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                    else:
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)

                except FileExistsError:
                    print("Directory path already exists, continue.")
        return added_docs

    # Gets all documents located at second Decisions link
    # Date checking is correct
    def get_docs_Decisions_v2(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        print("\n========================= Belgium Decisions_v2 Documents ===========================")

        host_link_input = "https://www.autoriteprotectiondonnees.be"
        start_path_input = "/citoyen/chercher?q=&search_category%5B%5D=taxonomy%3Apublications&search_type%5B%5D=decision&search_subtype%5B%5D=taxonomy%3Ageneral_secretary_international_decisions&search_subtype%5B%5D=taxonomy%3Ageneral_secretary_general_decisions&s=recent&l=25"
        iteration_number = 1
        page_url = host_link_input + start_path_input
        page_source = self.get_source(page_url)
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        pagination = self.update_pagination(page_soup=page_soup, host_link_input=host_link_input, start_path_input=start_path_input)

        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url)
            if page_source is None:
                continue
            page_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert page_soup
            search_result = page_soup.find('div', id='search-result')
            assert search_result
            # s1. Results
            for media in search_result.find_all('div', class_='media'):
                time.sleep(1)
                media_title = media.find('h3', class_='media-title')
                print("------------ Document " + str(iteration_number) + " ------------")
                iteration_number += 1
                # print('title:', media_title)
                assert media_title
                result_link = media_title.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                # if document_href.endswith('.pdf') is False:
                #    continue
                host = "https://www.autoriteprotectiondonnees.be"
                document_url = host + document_href
                print('document_url:', document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=1000)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Decisions 2' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    if document_url.endswith('.pdf') is False:
                        document_soup = BeautifulSoup(document_response.text, 'html.parser')
                        assert document_soup
                        date_text = document_soup.find('div', class_='date').get_text()
                        date_str = date_text[-4:]  # date_text[:-4] + ' ' + date_text[-4:]
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        date = datetime.date(tmp.year, tmp.month, tmp.day)

                        if date.year < 2018:
                            print("Skipping outdated document")
                            continue
                        page_body = document_soup.find('div', class_='page-body')
                        assert page_body
                        document_text = page_body.get_text()
                    else:
                        date_str = document_title.split(' du ')[-1]
                        print("date_str:", date_str)
                        tmp = dateparser.parse(date_str, languages=[self.language_code])

                        # Conditional checks for document date
                        if tmp is None:
                            media_date = media.find('span', class_="media-date")
                            assert media_date
                            year = int(media_date.get_text())
                            if year < 2018:
                                print("Skipping outdated document")
                                continue
                        else:
                            date = datetime.date(tmp.year, tmp.month, tmp.day)
                            #if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                                #continue
                            if date.year < 2018:
                                print("Skipping outdated document")
                                continue

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_response.content)
                    if document_url.endswith('.pdf'):
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                    else:
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print('Directory path already exists, continue.')
        return added_docs

    # Gets all documents located at opinions link
    # Date checking correct
    # Only visits first page, since rest of pages are all outdated (only call update pagination once)
    def get_docs_Opinions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Belgium Opinion Documents ===========================")
        added_docs = []

        host_link_input = "https://www.autoriteprotectiondonnees.be"
        start_path_input = "/citoyen/chercher?q=GDPR&search_category%5B%5D=taxonomy%3Apublications&search_type%5B%5D=advice&s=recent&l=50"
        iteration_number = 1
        page_url = host_link_input + start_path_input
        page_source = self.get_source(page_url)
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        pagination = self.update_pagination(page_soup=page_soup)
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url)
            if page_source is None:
                continue
            page_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert page_soup
            search_result = page_soup.find('div', id='search-result')
            assert search_result
            # s1. Results
            for media in search_result.find_all('div', class_='media'):
                time.sleep(1)
                media_title = media.find('h3', class_='media-title')
                print("------------ Document " + str(iteration_number) + " ------------")
                iteration_number += 1
                # print('title:', media_title)
                assert media_title
                result_link = media_title.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                # if document_href.endswith('.pdf') is False:
                #    continue
                host = "https://www.autoriteprotectiondonnees.be"
                document_url = host + document_href
                print('document_url:', document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=1000)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                dpa_folder = self.path

                document_folder = dpa_folder + '/' + 'Opinions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    if document_url.endswith('.pdf') is False:
                        document_soup = BeautifulSoup(document_response.text, 'html.parser')
                        assert document_soup
                        date_text = document_soup.find('div', class_='date').get_text()
                        date_str = date_text[-4:]  # date_text[:-4] + ' ' + date_text[-4:]
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        date = datetime.date(tmp.year, tmp.month, tmp.day)
                        if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                            return added_docs
                        if date.year < 2018:
                            print("Skipping outdated document")
                            return added_docs

                        page_body = document_soup.find('div', class_='page-body')
                        assert page_body
                        document_text = page_body.get_text()
                    else:
                        date_str = document_title.split(' du ')[-1]
                        print("date_str:", date_str)
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        if tmp is None:
                            media_date = media.find('span', class_="media-date")
                            assert media_date
                            year = int(media_date.get_text())
                            if year < 2018:
                                return added_docs
                        else:
                            date = datetime.date(tmp.year, tmp.month, tmp.day)
                            if date.year < 2018:
                                print("Skipping outdated document")
                                return added_docs

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_response.content)
                    if document_url.endswith('.pdf'):
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                    else:
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                        },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print('Directory path already exists, continue.')
        return added_docs

    # Gets all documents located at guides link
    def get_docs_Guides(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Belgium Guides ===========================")
        added_docs = []

        host_link_input = "https://www.autoriteprotectiondonnees.be"
        start_path_input = "/citoyen/chercher?q=&search_category%5B%5D=taxonomy%3Apublications&search_type%5B%5D=recommendation&s=recent&l=25"
        page_url = host_link_input + start_path_input
        page_source = self.get_source(page_url)
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        pagination = self.update_pagination(page_soup=page_soup)

        iteration_number = 1
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url)
            if page_source is None:
                continue
            page_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert page_soup
            search_result = page_soup.find('div', id='search-result')
            assert search_result
            # s1. Results
            for media in search_result.find_all('div', class_='media'):
                time.sleep(5)
                media_title = media.find('h3', class_='media-title')
                print("------------ Document " + str(iteration_number) + " ------------")
                iteration_number += 1
                # print('title:', media_title)
                assert media_title
                result_link = media_title.find('a')
                # s2. Documents
                document_title = result_link.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://www.autoriteprotectiondonnees.be"
                document_url = host + document_href
                print('document_url:', document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=1000)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Guides' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    if document_url.endswith('.pdf') is False:
                        document_soup = BeautifulSoup(document_response.text, 'html.parser')
                        assert document_soup
                        date_text = document_soup.find('div', class_='date').get_text()
                        date_str = date_text[-4:]  # date_text[:-4] + ' ' + date_text[-4:]
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        date = datetime.date(tmp.year, tmp.month, tmp.day)
                        if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                            print("Skipping outdated document")
                            return added_docs
                        if date.year < 2018:
                            print("Skipping outdated document")
                            return added_docs
                        page_body = document_soup.find('div', class_='page-body')
                        assert page_body
                        document_text = page_body.get_text()
                    else:
                        date_str = document_title.split(' du ')[-1]
                        tmp = dateparser.parse(date_str, languages=[self.language_code])
                        if tmp is None:
                            media_date = media.find('span', class_="media-date")
                            assert media_date
                            year = int(media_date.get_text())
                            if year < 2018:
                                print("---> SKIPPING DOCUMENT BECAUSE OF DATE <---")
                                return added_docs
                        else:
                            date = datetime.date(tmp.year, tmp.month, tmp.day)
                            if date.year < 2018:
                                print("Skipping outdated document")
                                return added_docs

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_response.content)
                    if document_url.endswith('.pdf'):
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                    else:
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print('Directory path already exists, continue.')

        return added_docs

    # Gets all documents located at annual report link
    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Belgium Annual Reports ===========================")
        added_docs = []

        page_url = "https://www.autoriteprotectiondonnees.be/citoyen/l-autorite/rapport-annuel"
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url)
        if page_source is None:
            sys.exit("Couldn't obtain page_source from page_url")
        page_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert page_soup
        page_content = page_soup.find('section', id='page-content')
        assert page_content
        page_body = page_content.find('div', class_='page-body')
        assert page_body

        iteration_number = 1
        for expanded in page_body.find_all('div', class_='collapse'):
            time.sleep(5)
            assert expanded
            result_link = expanded.find_all('a')
            assert result_link

            for link in result_link:
                document_href = link.get('href')
                assert document_href

                # If the document_href is not a pdf or zip file, its not relevant to our objective
                if not (document_href.endswith('.pdf') or document_href.endswith('.zip')):
                    continue

                # Get the year of the document by slicing document_href
                href_get_year = document_href[slice(-8, -4)]
                href_year_int = int(href_get_year)

                # If the document is older than 2018, skip it
                if href_year_int < 2018:
                    continue

                # Get document title by slicing it out of the href
                document_title = document_href[slice(-23, -4)]

                print('\n------------ Document ' + str(iteration_number) + '-------------')
                iteration_number += 1
                print('Document Title: ' + document_title)

                # If we don't want to overwrite documents and we already have this document_has in existing_docs,
                # skip the document
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                host = "https://www.autoriteprotectiondonnees.be"
                document_url = host + document_href

                print('document_url:', document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=1000)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass

                if document_response is None:
                    continue

                dpa_folder = self.path

                document_folder = dpa_folder + '/' + 'AnnualReports' + '/' + document_hash
                try:
                    os.makedirs(document_folder)
                    # If the link downloads a .zip file -> extract it, then iterate through the html files within it
                    # and store their text in one big text file. Store .txt file in document_folder
                    if document_url.endswith('.zip'):
                        with open(document_folder + '/' + self.language_code + '.zip', 'wb') as f:
                            f.write(document_response.content)
                        # Extract zip file
                        file_name = document_folder + '/' + self.language_code + '.zip'
                        with ZipFile(file_name, 'r') as zip:
                            print('\n--- ZIP FILE CONTENT ---')
                            os.chdir(document_folder)
                            zip.printdir()
                            zip.extractall()

                            print('\n--- CONVERTING .HTML TO .TXT ---')

                            # Iterate through the extracted zip folder -> concatenate the text within html files to one big
                            # text file
                            all_text_concatenated = ''
                            html_iteration = 1
                            for file in os.listdir(document_folder):
                                filename = os.fsdecode(file)
                                if 'Rapport annuel' in filename:
                                    for html_file in os.listdir(filename):
                                        os.chdir(filename)
                                        print(html_file)
                                        with open(html_file, 'r') as f:
                                            contents = f.read()
                                            html_soup = BeautifulSoup(contents, 'html.parser')
                                            assert html_soup
                                            html_body = html_soup.find('body')
                                            assert html_body

                                            all_text_concatenated = html_body.get_text() + all_text_concatenated

                                        os.chdir(document_folder)

                                    # Store the text file in the document folder for the link
                                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                                        f.write(str.encode(all_text_concatenated))
                                        html_iteration += 1

                    # document_url ends with '.pdf'
                    else:
                        with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                            f.write(document_response.content)

                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': href_year_int,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                    iteration_number += 1
                except FileExistsError:
                    print('Directory path already exists, continue.')

        return added_docs



import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import sys
import urllib3

class Portugal(DPA):
    def __init__(self, path=os.curdir):
        country_code='pt'
        super().__init__(country_code, path)

    # I add pgd=1 at end of default start_path string
    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        source = {
            "host": "https://www.cnpd.pt",
            "start_path": "/decisoes/historico-de-decisoes/?year=2022"
        }
        host = source['host']
        start_path = source['start_path']

        # Just add initial page
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)

            # add previous year links
            previous_year = ['2021', '2020', '2019', '2018']
            for i in range(len(previous_year)):
                 start_path = start_path[:-4] + previous_year[i]
                 print('previous_year: ', host + start_path)
                 pagination.add_item(host + start_path)
        # Add next page
        else:
            c_pagination = page_soup.find('div', class_='c-pagination')
            if c_pagination is not None:
                for a in c_pagination.find_all('a'):
                    page_href = a.get('href')
                    # Only add link if not in objects link list
                    if pagination.has_link('href'):
                        continue
                    else:
                        pagination.add_item(page_href)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url, verify=False)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Reports(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guidelines(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Portugal Decision Documents =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        added_docs = []
        pagination = self.update_pagination()
        iteration = 1
        while pagination.has_next():

            page_url = pagination.get_next()
            if to_print:
                print('\nPage: ', page_url)

            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                sys.exit('Page source is None')

            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            layout = results_soup.find('div', class_='layout')
            assert layout

            for c_card in layout.find_all('div', class_='c-card'):
                result_link = c_card.find('a')
                if result_link is None:
                    continue

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1

                document_title = result_link.find('div', 'c-card-header-medium')
                assert document_title

                document_title = document_title.get_text()
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                print('\tDocument Title: ' + document_title)

                document_date = document_title[-4:]
                print('\tDocument Date: ' + document_date)

                if int(document_date) < 2018:
                    print("Skipping outdated document")
                    continue

                document_href = result_link.get('href')

                # ignore one page with "content not found" issue
                ignore_list = ['121657', '121658']
                if document_href[-6:] in ignore_list:
                    print("\tcontent not found")
                    continue

                host = "https://www.cnpd.pt"
                document_url = host + document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, verify=False)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                print('\tDocument URL: ' + document_url)

                document_content = document_response.content
                dpa_folder = self.path

                document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(link_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            self.update_pagination(pagination=pagination, page_soup=results_soup)

        return added_docs

    def get_docs_Reports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Portugal Reports =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        added_docs = []

        iteration = 1

        page_url = 'https://www.cnpd.pt/cnpd/relatorios-de-atividades/'
        if to_print:
            print('\nPage: ', page_url)

        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit('Page source is None')

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        layout = results_soup.find('div', class_='layout')
        assert layout

        c_content = layout.find('div', class_='c-content-text')
        assert c_content

        for p_tag in c_content.find_all('p'):
            assert p_tag

            result_link = p_tag.find('a')
            if result_link is None:
                continue

            print('\n------------ Document ' + str(iteration) + ' ------------')
            iteration += 1

            document_title = result_link.get_text()
            assert document_title

            print('\tDocument Title:\t' + document_title)

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_date = document_title[-4:]
            print('\tDocument Date:\t' + document_date)

            if int(document_date) < 2018:
                print("\tSKIPPING OUTDATED DOCUMENT")
                continue

            document_href = result_link.get('href')
            host = "https://www.cnpd.pt"
            document_url = host + document_href

            if to_print:
                print("\tDocument:\t", document_hash)
            document_response = None
            try:
                document_response = requests.request('GET', document_url, verify=False)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            print('\tDocument URL:\t' + document_url)

            document_content = document_response.content
            dpa_folder = self.path

            document_folder = dpa_folder + '/' + 'Reports' + '/' + document_hash
            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_content)
                try:
                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        link_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(link_text)
                except:
                    print('Can not convert pdf to txt')
                    pass
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_date,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return added_docs

    def get_docs_Guidelines(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Portugal Guidelines =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        added_docs = []
        pagination = self.update_pagination()

        page_url = 'https://www.cnpd.pt/organizacoes/orientacoes-e-recomendacoes/'
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit('Page source is None')

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        layout = results_soup.find('div', class_='layout')
        assert layout

        side_nav = layout.find('ul', class_='sidenav-content')
        assert side_nav

        iteration = 1
        for li in side_nav.find_all('li'):
            assert li

            a_tag = li.find('a')
            assert a_tag

            href = a_tag.get('href')
            assert href

            if href.startswith('http') and 'download' in href:
                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1
                document_title = a_tag.get_text()
                print('\tdocument_title: ' + document_title)
                print('\tPDF Link: ' + href)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                print('\tdocument_hash: ' + document_hash)

                response = None
                try:
                    response = requests.request('GET', href, verify=False)
                    response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if response is None:
                    continue

                content = response.content

                dpa_folder = self.path
                document_folder = dpa_folder + '/' 'Guidelines' + '/' + document_hash

                try:
                    os.makedirs(document_folder)
                except FileExistsError:
                    pass
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(content)
                try:
                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        document_text = PDFToTextService().text_from_pdf_path(document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                except:
                    print('\tFailed to convert PDF to text')

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': None,
                        'url': href
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
                continue

            url = "https://www.cnpd.pt" + href

            response = None
            try:
                response = requests.request('GET', url, verify=False)
                response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if response is None:
                continue

            page_soup = BeautifulSoup(response.text, 'html.parser')
            assert page_soup

            page_layout = page_soup.find('div', class_='layout')
            assert page_layout

            c_content = page_layout.find('div', class_='c-content-text')
            assert c_content

            for p in c_content.find_all('p'):
                assert p

                a = p.find('a')
                assert a

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1

                document_title = a.get_text()
                print('\tDocument Title: ' + document_title)

                p_tag_all_text = p.get_text()
                assert p_tag_all_text

                document_date = p_tag_all_text[-4:]

                print('\tDocument Date: ' + document_date)

                if int(document_date) < 2018:
                    print('\tSkipping outdated document')
                    continue

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                pdf_href = a.get('href')
                assert pdf_href

                pdf_url = "https://www.cnpd.pt" + pdf_href

                print('\tPDF Link: ' + pdf_url)

                response = None
                try:
                    pdf_response = requests.request('GET', pdf_url, verify=False)
                    pdf_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if pdf_response is None:
                    continue

                pdf_content = pdf_response.content

                dpa_folder = self.path
                document_folder = dpa_folder + '/' 'Guidelines' + '/' + document_hash

                try:
                    os.makedirs(document_folder)
                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_content)
                    try:
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            document_text = PDFToTextService().text_from_pdf_path(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                    except:
                        print('\tFailed to convert PDF to text')

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_date,
                            'url': pdf_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return added_docs



import re
import nltk
nltk.download('averaged_perceptron_tagger')
nltk.download('wordnet')
#nltk.download('stopwords')
from nltk.corpus import stopwords
from nltk.stem.porter import PorterStemmer
from nltk.tokenize import RegexpTokenizer
#nltk.download('wordnet')
from nltk.stem.wordnet import WordNetLemmatizer
from sklearn.feature_extraction.text import CountVectorizer
from pygdpr.services.metadata_service import *

class KeywordsMetadataService(MetadataService):
    def for_text(self, text, preprocess=True, n_keywords=20, custom_stopwords=[]):
        _stopwords = set(stopwords.words("english")).union(custom_stopwords)
        text = re.sub('[^a-zA-Z]', ' ', text)
        text = text.lower()
        text = re.sub("&lt;/?.*?&gt;"," &lt;&gt; ", text)
        text = re.sub("(\\d|\\W)+", " ", text)
        text = text.split()
        lem = WordNetLemmatizer()
        text = [lem.lemmatize(word) for word in text if not word in _stopwords]
        tags = nltk.pos_tag(text)
        text = [word for word, pos in tags if (pos != 'RB')]
        corpus = ' '.join(text)
        vec = CountVectorizer(ngram_range=(1,3)).fit([corpus])
        bag_of_words = vec.transform([corpus])
        sum_words = bag_of_words.sum(axis=0)
        words_freq = [(word, sum_words[0, idx]) for word, idx in
                       vec.vocabulary_.items()]
        words_freq = sorted(words_freq, key = lambda x: x[1], reverse=True)
        language_code = 'en'
        return [{ "name": w, "language_code": language_code } for w, freq in words_freq[:n_keywords]]



import os
import math
import time
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
from selenium.common.exceptions import NoSuchElementException

class Denmark(DPA):
    def __init__(self, path=os.curdir):
        country_code='DK'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, results_soup=None, driver=None, start_path="Decisions"):
        source = {
            'host': 'https://www.datatilsynet.dk',
            'start_path_Decisions': '/afgoerelser/afgoerelser?page=1',
            "start_path_Permissions": "/afgoerelser/tilladelser?page=1"
        }
        host = source['host']
        if start_path != "Decisions":
            start_path = source['start_path_Permissions']
        else:
            start_path = source['start_path_Decisions']
        page_url = ""
        if pagination is None:
            page_url = host + start_path
        else:
            pagination = driver.find_element_by_class_name('pagination')
            try:
                item = pagination.find_element_by_class_name("next")
                if item is not None:
                    page_url = item.get_attribute("href")
                    assert page_url

                    # there is a bug in Denmark websides: the next page button has a wrong href link
                    # so we replace it to the correct link
                    current_page = page_url
                    page_string = current_page.split('&PageNumber=')
                    if page_string[0][-2].isdigit():
                        page_url = page_string[0][:-2] + page_string[1]
                    else:
                        page_url = page_string[0][:-1]+page_string[1]
            except NoSuchElementException:
                pass

        if len(page_url) != 0:
            exec_path = WebdriverExecPolicy().get_system_path()
            options = webdriver.ChromeOptions()
            options.add_argument('headless')
            driver = webdriver.Chrome(options=options, executable_path=exec_path)
            driver.get(page_url)
            pagination = Pagination()
            pagination.add_item(driver)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (driver is not None)
        try:
            WebDriverWait(driver, 10).until(EC.presence_of_element_located(
                (By.CLASS_NAME, 'archive-search-result')
            ))
            WebDriverWait(driver, 10).until(EC.presence_of_element_located(
                (By.CLASS_NAME, 'items')
            ))
        except:
            return None
        page_source = driver.page_source
        return page_source

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions_1(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Decisions_2(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Permissions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    # add all the decision documents
    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        added_docs += self.get_docs_Decisions_1(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Decisions_2(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Decisions_1(self, existing_docs=[], overwrite=False, to_print=True):
        existed_docs = []
        dict_hashcode = {}
        page_list = []
        print("\n========================= Denmark Decisions_1 Documents ===========================")
        # s0. Pagination
        try:
            pagination = self.update_pagination()
            while pagination.has_next():
                driver = pagination.get_next()
                page_number = ''
                if to_print:
                    current_url = driver.current_url
                    print('Page:\t', current_url)
                    for i in current_url:
                        if i.isdigit():
                            page_number += i

                if page_number in page_list:
                    continue
                #print('page_number: ', page_number)
                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                driver_doc.get(current_url)
                ajaxhost = driver.find_element_by_class_name("ajaxhost")
                if "No results found" in ajaxhost.text:
                    return existed_docs
                items = ajaxhost.find_element_by_class_name("items")
                item_list = items.find_elements_by_class_name("item")
                for i in range(len(item_list)):
                    # date
                    date_str = item_list[i].find_element_by_class_name('date').text
                    assert date_str
                    date_str = date_str.strip().split(' ')[-1]
                    #print('date_str: ', date_str)
                    tmp = datetime.datetime.strptime(date_str, '%d-%m-%Y')
                    date = datetime.date(tmp.year, tmp.month, tmp.day)
                    print("\tdate: ", date)
                    if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                        print('\tNo satisfied by date\t', date)
                        continue
                    # document
                    i += 1
                    document = items.find_element_by_xpath('//*[@id="ContentPlaceHolderDefault_searchResult"]/div[2]/div/div/div['+str(i)+']/div/h2/a')
                    assert document
                    document_url = document.get_attribute("href")
                    assert document_url
                    print("document_url: ", document_url)
                    document_title = document.text
                    document_hash = hashlib.md5(document_title.encode()).hexdigest()

                    if document_hash in existing_docs and overwrite == False:
                        if to_print:
                            print('\tSkipping existing document:\t', document_hash)
                        continue
                    if document_hash in dict_hashcode.keys() and dict_hashcode[document_hash] == date:
                        if date_str == '17-05-2018':
                            print ('\t Last document')
                            return existed_docs
                        print('\tSkipping existing document:\t', document_hash)
                        continue
                    if document_hash in dict_hashcode:
                        # documents have the same hashcode, but different dates
                        document_hash = document_hash + '-' + date_str
                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)
                        print('\tdocument_title: ', document_title)
                        exec_path = WebdriverExecPolicy().get_system_path()
                        options = webdriver.ChromeOptions()
                        options.add_argument('headless')
                        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                        driver_doc.get(document_url)
                        document_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
                        assert document_soup
                        WebDriverWait(driver_doc, 15).until(EC.presence_of_element_located(
                            (By.CLASS_NAME, 'news-page')
                        ))
                        #print("made it thus far to news_page")
                        news_page = document_soup.find('div', class_='news-page')
                        assert news_page
                        document_text = news_page.get_text()
                        document_text = document_text.lstrip()

                        #print("writing to the directory:", document_folder)
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date.strftime('%d/%m/%Y'),
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        dict_hashcode[document_hash] = date
                        existed_docs.append(document_hash)

                    except FileExistsError:
                        print("Directory path already exists, continue.")
                page_list.append(page_number)
                pagination = self.update_pagination(pagination=pagination, driver=driver)
        except AttributeError:
            print('no more documents')
        return existed_docs

    def get_docs_Decisions_2(self, existing_docs=[], overwrite=False, to_print=True):
        existed_docs = []
        source = {
            "host": "https://www.datatilsynet.dk",
            "start_path": "/afgoerelser/boedesager"
        }
        host = source['host']
        start_path = source['start_path']
        source_url = host + start_path
        results_response = requests.request('GET', source_url)
        results_content = results_response.content
        results_soup = BeautifulSoup(results_content, 'html.parser')
        web_page = results_soup.find('div', class_='web-page')
        rich_text = web_page.find('div', class_='rich-text')
        print("\n========================= Denmark Decisions_2 Documents ===========================")

        for li in rich_text.find('ul').find_all('li'):
            time.sleep(2)
            document_href = li.find('a').get('href')
            # print("document_href: ", document_href)
            document_url = host + document_href
            assert document_url
            exec_path = WebdriverExecPolicy().get_system_path()
            options = webdriver.ChromeOptions()
            options.add_argument('headless')
            driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
            driver_doc.get(document_url)
            content = driver_doc.find_element_by_xpath('//*[@id="wrapper"]/div/div[3]/div/div/div[2]/div/div[1]/div')
            if content is None:
                print('\tno content')
                continue
            document_title = content.find_element_by_class_name('heading')
            document_title = document_title.text.strip()
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite == False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Decisions_2' + '/' + document_hash
            try:
                os.makedirs(document_folder)
                print("document_title: ", document_title)
                print("\tdocument_hash: ", document_hash)
                date_str = content.find_element_by_class_name('date').text
                date_str = date_str.strip().split(' ')[-1]
                tmp = datetime.datetime.strptime(date_str, '%d-%m-%Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    print('\tNo satisfied by date\t', date)
                    continue
                print("\tdocument_url: ", document_url)

                document_text = content.text
                with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                    f.write(document_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': date.strftime('%d/%m/%Y'),
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)
            except FileExistsError:
                print('Directory path already exists, continue.')
        return existed_docs


    def get_docs_Permissions(self, existing_docs=[], overwrite=False, to_print=True):
        existed_docs = []
        dict_hashcode = {}
        page_list = []
        print("\n========================= Denmark Permissions Documents ===========================")
        # s0. Pagination
        pagination = self.update_pagination(start_path = "Permissions")
        try:
            while pagination.has_next():
                driver = pagination.get_next()
                page_number = ''
                if to_print:
                    current_url = driver.current_url
                    print('Page:\t', current_url)
                    for i in current_url:
                        if i.isdigit():
                            page_number += i
                if page_number in page_list:
                    continue
                #print('page_number: ', page_number)
                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                driver_doc.get(current_url)
                ajaxhost = driver.find_element_by_class_name("ajaxhost")
                if "No results found" in ajaxhost.text:
                    return existed_docs
                items = ajaxhost.find_element_by_class_name("items")
                item_list = items.find_elements_by_class_name("item")
                # print('len of item: ', len(item_list))
                for i in range(len(item_list)):
                    # date
                    date_str = item_list[i].find_element_by_class_name('date').text
                    assert date_str
                    date_str = date_str.strip().split(' ')[-1]
                    # print('date_str: ', date_str)
                    tmp = datetime.datetime.strptime(date_str, '%d-%m-%Y')
                    date = datetime.date(tmp.year, tmp.month, tmp.day)
                    # print("\tdate: ", date)
                    if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                        print('\tNot satisfied by date:\t', date)
                        continue
                    # document
                    i += 1
                    document = items.find_element_by_xpath('//*[@id="ContentPlaceHolderDefault_searchResult"]/div[2]/div/div/div['+str(i)+']/div/h2/a')
                    assert document
                    document_url = document.get_attribute("href")
                    assert document_url
                    print("document_url: ", document_url)
                    document_title = document.text
                    document_hash = hashlib.md5(document_title.encode()).hexdigest()

                    if document_hash in existing_docs and overwrite == False:
                        if to_print:
                            print('\tSkipping existing document:\t', document_title)
                        continue
                    if document_hash in dict_hashcode and dict_hashcode[document_hash] == date:
                        print('\tSkipping existing document:\t', document_title)
                        continue
                    if document_hash in dict_hashcode:
                        # documents have the same hashcode, but different dates
                        document_hash = document_hash + '-' + date_str

                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'Permissions' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)
                        print('\tdocument_title: ', document_title)
                        exec_path = WebdriverExecPolicy().get_system_path()
                        options = webdriver.ChromeOptions()
                        options.add_argument('headless')
                        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                        driver_doc.get(document_url)
                        document_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
                        assert document_soup
                        WebDriverWait(driver_doc, 15).until(EC.presence_of_element_located(
                            (By.CLASS_NAME, 'news-page')
                        ))
                        #print("made it thus far to news_page")
                        news_page = document_soup.find('div', class_='news-page')
                        assert news_page
                        document_text = news_page.get_text()
                        document_text = document_text.lstrip()

                        #print("writing to the directory:", document_folder)
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            f.write(document_text)
                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date.strftime('%d/%m/%Y'),
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        existed_docs.append(document_hash)
                        dict_hashcode[document_hash] = date
                    except FileExistsError:
                        print("Directory path already exists, continue.")
                page_list.append(page_number)
                pagination = self.update_pagination(pagination=pagination, driver=driver, start_path="Permissions")
        except AttributeError:
            print('no more documents')
        return existed_docs

    # Method not complete, should implement later
    def get_docs_Guides(self, existing_docs=[], overwrite=False, to_print=True):
        existed_docs = []
        dict_hashcode = {}
        page_list = []
        source = {
            'host': 'https://www.datatilsynet.dk',
            'start_path': '/hvad-siger-reglerne/vejledning'
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
        driver_doc.get(page_url)
        document_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
        assert document_soup
        WebDriverWait(driver_doc, 15).until(EC.presence_of_element_located(
            (By.CLASS_NAME, 'container')
        ))
        for div in document_soup.find_all('div', class_='span-4'):
            document = div.find('h2', class_='heading')
            document_href = document.find('a').get('href')
            document_url = host + document_href
            print('document_href: ', document_url)
            document_title = document.find('a').get_text()
            print('document_title: ', document_title)

            exec_path = WebdriverExecPolicy().get_system_path()
            options = webdriver.ChromeOptions()
            options.add_argument('headless')
            driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
            driver_doc.get(document_url)
            article_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
            assert article_soup
            WebDriverWait(driver_doc, 15).until(EC.presence_of_element_located(
                (By.CLASS_NAME, 'container')
            ))

            for div in article_soup.find_all('div', class_='module multi-box bg-color-c card'):
                text = div.find('div', class_='text')
                article_href = text.find('a').get('href')
                article_url = host + article_href
                article_title = text.find('a').get_text()
                print('\tarticle_url\t ', article_url)
                print('\tarticle_title\t ', article_title)
                article_hash = hashlib.md5(article_title.encode()).hexdigest()
                # add date
                if article_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', article_title)
                    continue
                #if article_hash in dict_hashcode and dict_hashcode[article_hash] == date:
                #    print('\tSkipping existing document:\t', document_title)
                #    continue
        return existed_docs

    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Denmark Annual Reports ===========================")
        existed_docs = []
        source = {
            'host': 'https://www.datatilsynet.dk',
            'start_path': '/om-datatilsynet/aarsberetninger-og-aarsrapporter'
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
        driver_doc.get(page_url)
        document_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
        assert document_soup
        WebDriverWait(driver_doc, 15).until(EC.presence_of_element_located(
            (By.CLASS_NAME, 'container')
        ))
        for div in document_soup.find_all('div', class_='module link-list card drop-down'):
            for option in div.find_all('option'):
                text = option.get_text().strip('\n')
                year = option.get_text()
                if not text.isdigit():
                    continue
                elif int(year) < 2018:
                    continue
                #print('year: ', year)
                article_href = option.get('data-href')
                article_url = host + article_href
                print('article_url: ', article_url)
                article_title = article_href.split('/')[-1]
                print('article_title: ', article_title)
                article_hash = hashlib.md5(article_title.encode()).hexdigest()
                if article_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_title)
                try:
                    article_response = requests.request('GET', article_url)
                    article_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + article_hash
                # document_folder = dpa_folder + '/denmark' + '/' + 'Annual Reports' + '/' + article_hash
                try:
                    os.makedirs(document_folder)
                    if article_response is None:
                        continue
                    article_content = article_response.content
                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(article_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        document_text = PDFToTextService().text_from_pdf_path(
                            document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: article_title
                            },
                            'md5': article_hash,
                            'releaseDate': year,
                            'url': article_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(article_hash)
                except FileExistsError:
                    print("Directory path already exists, continue.")
        return existed_docs



import os
from collections import Counter
import json
from .dpa_stopwords_service import DPAStopwordsService

class GDPRStopwordsService():
    def __init__(self, path):
        self.path = path
        f = open(os.path.abspath("gdpr/assets/eu-members.json"), 'r')
        self.eu_members = json.load(f)
        f.close()

    def get(self, n=35, to_save=True):
        gdpr_stopwords = {}
        gdpr_counter = Counter()
        for code in self.eu_members.keys():
            dpa_stopwords = DPAStopwordsService(path=self.path).for_code(code, n=n, to_save=to_save)
            dpa_counter = Counter(dpa_stopwords)
            gdpr_counter = gdpr_counter + dpa_counter
        gdpr_stopwords = dict(gdpr_counter.most_common(n))
        if to_save:
            with open(os.path.abspath('gdpr/assets/gdpr-stopwords.txt'), 'w') as outfile:
                outfile.write('\n'.join(gdpr_stopwords.keys()))
            with open(os.path.abspath('gdpr/assets/gdpr-stopwords.json'), 'w') as outfile:
                json.dump(gdpr_stopwords, outfile, indent=4)
        return gdpr_stopwords



class NotReachedDailyTranslateQuotaSpec():
    def __init__(self, quota_service):
        self.quota_service = quota_service
    def is_satisfied_by(self, quota):
        return (quota <= self.quota_service.get_chars_per_day() or\
                self.quota_service.get_chars_per_day() == -1) # -1 == Unlimited value



from pygdpr.models.document.metadata.title import *
from pygdpr.models.document.metadata.summary import *

class Metadata(object):
    def __init__(self, titles, summaries, md5, release_date, release_year, release_month, source_url):
        self.titles = titles
        self.summaries = summaries
        self.md5 = md5
        self.release_date = release_date
        self.release_year = release_year
        self.release_month = release_month
        self.source_url = source_url



class PyLineInspectorService():
    # subroutines
    def should_skip_line(self, i, lines):
        line = lines[i]
        return (i >= 0 and i <= len(lines) - 2) and line == '\n'

    def matching_start_line(self, line, type_, name):
        return line.lstrip(' ').startswith(type_ + ' ' + name)

    def matching_end_line(self, start_line, i, indentation):
        return indentation <= start_line[1] and i != start_line[0]-1

    def start_line_not_found(self, start_line):
        return start_line == (-1, -1)

    # main func
    def for_type_and_name(self, path, type_, name):
        if path.endswith('.py') is False:
            raise ValueError('Path is not a python file (.py):', path)

        start_line = (-1, -1) # (line, indentation)
        end_line = (-1, -1)

        with open(path, 'r') as f:
            lines = f.readlines()
            lines.append('') # Atom IDE approach is the same.
            for i in range(len(lines)):
                line = lines[i]
                indentation = 0

                for j in range(len(line)):
                    char = line[j]
                    if char == ' ':
                        indentation += 1
                    else:
                        break

                if self.should_skip_line(i, lines) is True:
                    continue

                # print('line %s - indentation:\t%d' % (i+1, indentation))
                if self.matching_start_line(line, type_, name) is True:
                    start_line = (i+1, indentation)

                # first, find start_line
                if self.start_line_not_found(start_line) is True:
                    continue

                # second, find end_line
                if self.matching_end_line(start_line, i, indentation) is True:
                    end_line = (i, indentation)
                    break

                #if line == lines[len(lines)-1] and (start_line != (-1, -1) and end_line == (-1, -1)):
                #    end_line = (i, indentation)
                #    break

        return (start_line[0], end_line[0])
